<?php namespace Myriad\NotifyClient\Utilities;

use ArrayAccess;
use Carbon\Carbon;
use Illuminate\Contracts\Support\Arrayable;
use Illuminate\Contracts\Support\Jsonable;
use Illuminate\Database\Eloquent\Concerns\GuardsAttributes;
use Illuminate\Database\Eloquent\Concerns\HasAttributes;
use Illuminate\Database\Eloquent\Concerns\HidesAttributes;
use Illuminate\Database\Eloquent\MassAssignmentException;
use Illuminate\Database\Eloquent\JsonEncodingException;
use JsonSerializable;

abstract class BasicModel implements ArrayAccess, Arrayable, Jsonable, JsonSerializable {

  use HasAttributes, HidesAttributes, GuardsAttributes;

  /**
   * Create a new model instance.
   *
   * @param  array $attributes
   */
  public function __construct(array $attributes = []) {
    $this->syncOriginal();
    $this->fill($attributes);
  }

  /**
   * Fill the model with an array of attributes.
   *
   * @param  array $attributes
   * @return $this
   * @throws \Exception
   */
  public function fill(array $attributes) {
    $totallyGuarded = $this->totallyGuarded();

    foreach ($this->fillableFromArray($attributes) as $key => $value) {
      // The developers may choose to place some attributes in the "fillable" array
      // which means only those attributes may be set through mass assignment to
      // the model, and all others will just get ignored for security reasons.
      if ($this->isFillable($key)) {
        $this->setAttribute($key, $value);
      } elseif ($totallyGuarded) {
        throw new MassAssignmentException($key);
      }
    }

    return $this;
  }

  /**
   * Fill the model with an array of attributes. Force mass assignment.
   *
   * @param  array $attributes
   * @return $this
   */
  public function forceFill(array $attributes) {
    return static::unguarded(
      function () use ($attributes) {
        return $this->fill($attributes);
      }
    );
  }

  /**
   * Fill the model with an array of attributes. Force mass assignment.
   *
   * @param  array $attributes
   * @return $this
   */
  public function hydrate(array $attributes) {
    $this->forceFill($attributes);
    $this->syncOriginal();
    return $this;
  }

  /**
   * Convert the model instance to an array.
   *
   * @return array
   */
  public function toArray() {
    return $this->attributesToArray();
  }

  /**
   * Convert the model instance to JSON.
   *
   * @param  int $options
   * @return string
   * @throws \Illuminate\Database\Eloquent\JsonEncodingException
   */
  public function toJson($options = 0) {
    $json = json_encode($this->jsonSerialize(), $options);

    if (JSON_ERROR_NONE !== json_last_error()) {
      throw JsonEncodingException::forModel($this, json_last_error_msg());
    }

    return $json;
  }

  /**
   * Convert the object into something JSON serializable.
   *
   * @return array
   */
  public function jsonSerialize() {
    return $this->toArray();
  }

  /**
   * Dynamically retrieve attributes on the model.
   *
   * @param  string $key
   * @return mixed
   */
  public function __get($key) {
    return $this->getAttribute($key);
  }

  /**
   * Dynamically set attributes on the model.
   *
   * @param  string $key
   * @param  mixed $value
   * @return void
   */
  public function __set($key, $value) {
    $this->setAttribute($key, $value);
  }

  /**
   * Determine if the given attribute exists.
   *
   * @param  mixed $offset
   * @return bool
   */
  public function offsetExists($offset) {
    return isset($this->$offset);
  }

  /**
   * Get the value for a given offset.
   *
   * @param  mixed $offset
   * @return mixed
   */
  public function offsetGet($offset) {
    return $this->$offset;
  }

  /**
   * Set the value for a given offset.
   *
   * @param  mixed $offset
   * @param  mixed $value
   * @return void
   */
  public function offsetSet($offset, $value) {
    $this->$offset = $value;
  }

  /**
   * Unset the value for a given offset.
   *
   * @param  mixed $offset
   * @return void
   */
  public function offsetUnset($offset) {
    unset($this->$offset);
  }

  /**
   * Determine if an attribute or relation exists on the model.
   *
   * @param  string $key
   * @return bool
   */
  public function __isset($key) {
    return !is_null($this->getAttribute($key));
  }

  /**
   * Unset an attribute on the model.
   *
   * @param  string $key
   * @return void
   */
  public function __unset($key) {
    unset($this->attributes[$key]);
  }

  /**
   * Handle dynamic static method calls into the method.
   *
   * @param  string $method
   * @param  array $parameters
   * @return mixed
   */
  public static function __callStatic($method, $parameters) {
    return (new static)->$method(...$parameters);
  }

  /**
   * Convert the model to its string representation.
   *
   * @return string
   */
  public function __toString() {
    return $this->toJson();
  }

  public function relationsToArray() {
    return [];
  }

  /**
   * Get a relationship.
   *
   * @param  string $key
   * @return mixed
   */
  public function getRelationValue($key) {
    return;
  }

  /**
   * Get the attributes that should be converted to dates.
   *
   * @return array
   */
  public function getDates() {
    return $this->dates;
  }

  /**
   * Get the casts array.
   *
   * @return array
   */
  public function getCasts() {
    return $this->casts;
  }

  /**
   * Get the format for database stored dates.
   *
   * @return string
   */
  protected function getDateFormat() {
    return Carbon::RFC3339;
  }

  /**
   * Create a new instance of the given model.
   *
   * @param  array $attributes
   * @return static
   */
  public function newInstance($attributes = []) {
    return new static((array)$attributes);
  }

  /**
   * Check if the model has a given attribute
   *
   * @param $key
   * @return bool
   */
  public function hasAttribute($key) {
    return array_key_exists($key, $this->attributes) || $this->hasGetMutator($key);
  }

}